@@ -1,5 +1,5 @@
 package Cassandra::Client::Metadata;
-use 5.008;
+use 5.010;
 use strict;
 use warnings;
 
@@ -24,6 +24,11 @@ sub add_prepared {
         decoder => $decoder,
         encoder => $encoder
     };
+    if (values %{$self->{prepare_cache}} > 500) {
+        unless ($self->{warned}++) {
+            warn "Cassandra::Client: found more than 500 queries in our prepared statement cache, try using placeholders";
+        }
+    }
     return;
 }
 
